# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import importlib.resources

from ..datamodule import CachingDataModule
from ..shenzhen.datamodule import RawDataLoader
from ..split import JSONDatabaseSplit


class DataModule(CachingDataModule):
    """Indian collection dataset for computer-aided diagnosis.

    The Indian collection database has been established to foster research
    in computer-aided diagnosis of pulmonary diseases with a special
    focus on pulmonary tuberculosis (TB).

    * Original resolution (height x width or width x height): more than 1024 x 1024
    * Split reference: [INDIAN-2013]_ with 20% of train set for the validation set

    Data specifications:

    * Raw data input (on disk):

        * PNG images (grayscale, encoded as RGB images with "inverted" grayscale scale)
        * Variable width and height

    * Output image:

        * Transforms:

            * Load raw PNG with :py:mod:`PIL`
            * Remove black borders
            * Torch center cropping to get square image

        * Final specifications:

            * Grayscale, encoded as a single plane image, 8 bits
            * Square, with varying resolutions, depending on the input image
    """

    def __init__(self, split_filename: str):
        super().__init__(
            database_split=JSONDatabaseSplit(
                importlib.resources.files(__name__.rsplit(".", 1)[0]).joinpath(
                    split_filename
                )
            ),
            raw_data_loader=RawDataLoader(),
        )
