# Copyright © 2022 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Aggregated dataset composed of Montgomery and Shenzhen datasets (cross
validation fold 6)"""

from clapper.logging import setup
from torch.utils.data.dataset import ConcatDataset

from .. import return_subsets
from ..base_datamodule import BaseDataModule, get_dataset_from_module
from ..montgomery.fold_6 import datamodule as mc_datamodule
from ..shenzhen.fold_6 import datamodule as ch_datamodule

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


class DefaultModule(BaseDataModule):
    def __init__(
        self,
        train_batch_size=1,
        predict_batch_size=1,
        drop_incomplete_batch=False,
        multiproc_kwargs=None,
    ):
        self.train_batch_size = train_batch_size
        self.predict_batch_size = predict_batch_size
        self.drop_incomplete_batch = drop_incomplete_batch
        self.multiproc_kwargs = multiproc_kwargs

        super().__init__(
            train_batch_size=train_batch_size,
            predict_batch_size=predict_batch_size,
            drop_incomplete_batch=drop_incomplete_batch,
            multiproc_kwargs=multiproc_kwargs,
        )

    def setup(self, stage: str):
        # Instantiate other datamodules and get their datasets
        module_args = {
            "train_batch_size": self.train_batch_size,
            "predict_batch_size": self.predict_batch_size,
            "drop_incomplete_batch": self.drop_incomplete_batch,
            "multiproc_kwargs": self.multiproc_kwargs,
        }

        mc = get_dataset_from_module(mc_datamodule, stage, **module_args)
        ch = get_dataset_from_module(ch_datamodule, stage, **module_args)

        # Combine datasets
        self.dataset = {}
        self.dataset["__train__"] = ConcatDataset(
            [mc["__train__"], ch["__train__"]]
        )
        self.dataset["train"] = ConcatDataset([mc["train"], ch["train"]])
        self.dataset["__valid__"] = ConcatDataset(
            [mc["__valid__"], ch["__valid__"]]
        )
        self.dataset["validation"] = ConcatDataset(
            [mc["validation"], ch["validation"]]
        )
        self.dataset["test"] = ConcatDataset([mc["test"], ch["test"]])

        (
            self.train_dataset,
            self.validation_dataset,
            self.extra_validation_datasets,
            self.predict_dataset,
        ) = return_subsets(self.dataset)


datamodule = DefaultModule
