# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""NIH CXR14 (relabeled) dataset for computer-aided diagnosis.

This dataset was extracted from the clinical PACS database at the National
Institutes of Health Clinical Center (USA) and represents 60% of all
their radiographs. It contains labels for fourteen common radiological
signs in this order: cardiomegaly, emphysema, effusion, hernia, infiltration,
mass, nodule, atelectasis, pneumothorax, pleural thickening, pneumonia,
fibrosis, edema and consolidation.
This is the relabeled version created in the CheXNeXt study.

* Reference: [NIH-CXR14-2017]_
* Original resolution (height x width or width x height): 1024 x 1024
* Labels: [CHEXNEXT-2018]_
* Split reference: [CHEXNEXT-2018]_
* Protocol ``default``:

  * Training samples: 98'637 (including labels)
  * Validation samples: 6'350 (including labels)
  * Test samples: 0

* Protocol `ìdiap``:
  * Images path adapted to Idiap infrastructure
"""

import importlib.resources
import os

from ...utils.rc import load_rc
from ..dataset import JSONDataset
from ..loader import load_pil_rgb, make_delayed

_protocols = [
    importlib.resources.files(__name__).joinpath("default.json.bz2"),
    importlib.resources.files(__name__).joinpath("cardiomegaly.json.bz2"),
]

_datadir = load_rc().get("datadir.nih_cxr14_re", os.path.realpath(os.curdir))
_idiap_folders = load_rc().get("nih_cxr14_re.idiap_folder_structure", False)


def _raw_data_loader(sample):
    if _idiap_folders:
        # for folder lookup efficiency, data is split into subfolders
        # each original file is on the subfolder `f[:5]/f`, where f
        # is the original file basename
        basename = os.path.basename(sample["data"])
        return dict(
            data=load_pil_rgb(
                os.path.join(
                    _datadir,
                    os.path.dirname(sample["data"]),
                    basename[:5],
                    basename,
                )
            ),
            label=sample["label"],
        )
    else:
        return dict(
            data=load_pil_rgb(os.path.join(_datadir, sample["data"])),
            label=sample["label"],
        )


def _loader(context, sample):
    # "context" is ignored in this case - database is homogeneous
    # we returned delayed samples to avoid loading all images at once
    return make_delayed(sample, _raw_data_loader)


json_dataset = JSONDataset(
    protocols=_protocols,
    fieldnames=("data", "label"),
    loader=_loader,
)
"""NIH CXR14 (relabeled) dataset object."""


def _maker(protocol, size=512):
    import torchvision.transforms as transforms

    from .. import make_dataset

    # ImageNet normalization
    normalize = transforms.Normalize(
        mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
    )

    return make_dataset(
        [json_dataset.subsets(protocol)],
        [transforms.Resize((size, size))],
        [transforms.RandomHorizontalFlip()],
        [transforms.ToTensor(), normalize],
    )
