# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""TBX11K simplified dataset for computer-aided diagnosis.

The TBX11K database has been established to foster research
in computer-aided diagnosis of pulmonary diseases with a special
focus on tuberculosis (aTB). The dataset was specifically
designed to be used with CNNs. It contains 11,000 chest X-ray
images, each of a unique patient. They were labeled by expert
radiologists with 5 - 10+ years of experience. Possible labels
are: "healthy", "active TB", "latent TB", and "sick & non-tb".
The version of the dataset used in this benchmark is a simplified.

* Reference: [TBX11K-SIMPLIFIED-2020]_
* Original (released) resolution (height x width or width x height): 512 x 512
* Split reference: none
* Protocol ``default``:

  * Training samples: 62.6% of CXR (including labels)
  * Validation samples: 16% of CXR (including labels)
  * Test samples: 21.4% of CXR (including labels)
"""

import importlib.resources
import os

from ...utils.rc import load_rc
from ..dataset import JSONDataset
from ..loader import load_pil_grayscale, make_delayed, make_delayed_bbox

_protocols = [
    importlib.resources.files(__name__).joinpath("default.json.bz2"),
    importlib.resources.files(__name__).joinpath("fold_0.json.bz2"),
    importlib.resources.files(__name__).joinpath("fold_1.json.bz2"),
    importlib.resources.files(__name__).joinpath("fold_2.json.bz2"),
    importlib.resources.files(__name__).joinpath("fold_3.json.bz2"),
    importlib.resources.files(__name__).joinpath("fold_4.json.bz2"),
    importlib.resources.files(__name__).joinpath("fold_5.json.bz2"),
    importlib.resources.files(__name__).joinpath("fold_6.json.bz2"),
    importlib.resources.files(__name__).joinpath("fold_7.json.bz2"),
    importlib.resources.files(__name__).joinpath("fold_8.json.bz2"),
    importlib.resources.files(__name__).joinpath("fold_9.json.bz2"),
]

_datadir = load_rc().get(
    "datadir.tbx11k_simplified", os.path.realpath(os.curdir)
)


def _raw_data_loader(sample):
    return dict(
        data=load_pil_grayscale(os.path.join(_datadir, sample["data"])),  # type: ignore
        label=sample["label"],
    )


def _raw_data_loader_bbox(sample):
    return dict(
        data=load_pil_grayscale(os.path.join(_datadir, sample["data"])),  # type: ignore
        label=sample["label"],
        bboxes=sample["bboxes"],
    )


def _loader(context, sample):
    # "context" is ignored in this case - database is homogeneous
    # we return delayed samples to avoid loading all images at once
    return make_delayed(sample, _raw_data_loader)


def _loader_bbox(context, sample):
    # "context" is ignored in this case - database is homogeneous
    # we return delayed samples to avoid loading all images at once
    return make_delayed_bbox(sample, _raw_data_loader_bbox)


json_dataset = JSONDataset(
    protocols=_protocols,
    fieldnames=("data", "label"),
    loader=_loader,
)

json_dataset_with_bboxes = JSONDataset(
    protocols=_protocols,
    fieldnames=("data", "label", "bboxes"),
    loader=_loader_bbox,
)
"""TBX11K simplified dataset object."""


def _maker(protocol, RGB=False):
    from torchvision import transforms

    from .. import make_dataset
    from ..augmentations import ElasticDeformation

    post_transforms = []
    if RGB:
        post_transforms = [
            transforms.Lambda(lambda x: x.convert("RGB")),
            transforms.ToTensor(),
        ]

    return make_dataset(
        [json_dataset.subsets(protocol)],
        [],
        [ElasticDeformation(p=0.8)],
        post_transforms,
    )
