# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import click

from clapper.click import verbosity_option
from clapper.logging import setup

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


@click.command(
    epilog="""Examples:

\b
  1. Aggregate multiple predictions csv files into one:

     .. code:: sh

        ptbench aggregpred -vv path/to/train/predictions.csv path/to/test/predictions.csv
""",
)
@click.argument(
    "label_path",
    nargs=-1,
)
@click.option(
    "--output-folder",
    "-f",
    help="Path where to store the aggregated csv file (created if necessary)",
    required=False,
    default=None,
    type=click.Path(dir_okay=True, file_okay=False),
)
@verbosity_option(logger=logger, expose_value=False)
def aggregpred(label_path, output_folder) -> None:
    """Aggregate multiple predictions csv files into one."""
    import os
    import re
    import shutil

    import pandas
    import torch

    # loads all data
    series = []
    for predictions_path in label_path:
        # Load predictions
        logger.info(f"Loading predictions from {predictions_path}...")
        pred_data = pandas.read_csv(predictions_path)
        pred = (
            torch.Tensor(
                [
                    eval(
                        re.sub(" +", " ", x.replace("\n", "")).replace(" ", ",")
                    )
                    for x in pred_data["likelihood"].values
                ]
            )
            .double()
            .flatten()
        )
        gt = (
            torch.Tensor(
                [
                    eval(
                        re.sub(" +", " ", x.replace("\n", "")).replace(" ", ",")
                    )
                    for x in pred_data["ground_truth"].values
                ]
            )
            .double()
            .flatten()
        )

        pred_data["likelihood"] = pred
        pred_data["ground_truth"] = gt

        series.append(pred_data)

    df = pandas.concat([s for s in series])

    logger.info(f"Output folder: {output_folder}")
    os.makedirs(output_folder, exist_ok=True)

    output_file = os.path.join(output_folder, "aggregpred.csv")
    if os.path.exists(output_file):
        backup = output_file + "~"
        if os.path.exists(backup):
            os.unlink(backup)
        shutil.move(output_file, backup)

    logger.info("Saving aggregated CSV file...")
    df.to_csv(output_file, index=False, header=True)
