# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os

from collections import defaultdict

import click

from clapper.click import ConfigCommand, ResourceOption, verbosity_option
from clapper.logging import setup
from matplotlib.backends.backend_pdf import PdfPages

from ..data.datamodule import ConcatDataModule
from ..data.typing import DataLoader
from ..utils.plot import precision_recall_f1iso, roc_curve
from ..utils.table import performance_table

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


def _validate_threshold(
    threshold: int | float | str, dataloader_dict: dict[str, DataLoader]
):
    """Validates the user threshold selection.

    Parameters
    ----------

    threshold:
        This number is used to define positives and negatives from
        probability maps, and report F1-scores (a priori). It
        should either come from the training set or a separate validation set
        to avoid biasing the analysis.  Optionally, if you provide a multi-set
        dataset as input, this may also be the name of an existing set from
        which the threshold will be estimated (highest F1-score) and then
        applied to the subsequent sets.  This number is also used to print
        the test set F1-score a priori performance

    dataloader_dict:
        Dictionary of set_name: dataloader, there set_name is the name of a dataset split
        and dataloader is the torch dataloader for that split.

    Returns
    -------

        The parsed threshold.
    """
    if threshold is None:
        return 0.5

    try:
        # we try to convert it to float first
        threshold = float(threshold)
        if threshold < 0.0 or threshold > 1.0:
            raise ValueError("Float thresholds must be within range [0.0, 1.0]")
    except ValueError:
        # it is a bit of text - assert dataset with name is available
        if not isinstance(dataloader_dict, dict):
            raise ValueError(
                "Threshold should be a floating-point number "
                "if your provide only a single dataset for evaluation"
            )
        if threshold not in dataloader_dict:
            raise ValueError(
                f"Text thresholds should match dataset names, "
                f"but {threshold} is not available among the datasets provided ("
                f"({', '.join(dataloader_dict.keys())})"
            )

    return threshold


@click.command(
    entry_point_group="ptbench.config",
    cls=ConfigCommand,
    epilog="""Examples:

\b
    1. Runs evaluation on an existing dataset configuration:

       .. code:: sh

          ptbench evaluate -vv montgomery --predictions-folder=path/to/predictions --output-folder=path/to/results
""",
)
@click.option(
    "--output-folder",
    "-o",
    help="Path where to store the analysis result (created if does not exist)",
    required=True,
    default="results",
    type=click.Path(),
    cls=ResourceOption,
)
@click.option(
    "--predictions-folder",
    "-p",
    help="Path where predictions are currently stored",
    required=True,
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    cls=ResourceOption,
)
@click.option(
    "--datamodule",
    "-d",
    help="A lighting data module containing the training and validation sets.",
    required=True,
    cls=ResourceOption,
)
@click.option(
    "--threshold",
    "-t",
    help="This number is used to define positives and negatives from "
    "probability maps, and report F1-scores (a priori). It "
    "should either come from the training set or a separate validation set "
    "to avoid biasing the analysis.  Optionally, if you provide a multi-set "
    "dataset as input, this may also be the name of an existing set from "
    "which the threshold will be estimated (highest F1-score) and then "
    "applied to the subsequent sets.  This number is also used to print "
    "the test set F1-score a priori performance",
    default=None,
    show_default=False,
    required=True,
    cls=ResourceOption,
)
@click.option(
    "--steps",
    "-S",
    help="This number is used to define the number of threshold steps to "
    "consider when evaluating the highest possible F1-score on test data.",
    default=1000,
    show_default=True,
    required=True,
    cls=ResourceOption,
)
@verbosity_option(logger=logger, cls=ResourceOption, expose_value=False)
def evaluate(
    output_folder: str,
    predictions_folder: str,
    datamodule: ConcatDataModule,
    threshold: int | float | str,
    steps: int,
    **_,
) -> None:
    """Evaluates a CNN on a tuberculosis prediction task.

    Note: batch size of 1 is required on the predictions.
    """

    from ..engine.evaluator import run

    datamodule.set_chunk_size(1, 1)
    datamodule.model_transforms = []

    datamodule.prepare_data()
    datamodule.setup(stage="predict")

    dataloader = datamodule.predict_dataloader()

    threshold = _validate_threshold(threshold, dataloader)

    if isinstance(threshold, str):
        # first run evaluation for reference dataset
        logger.info(f"Evaluating threshold on '{threshold}' set")
        _, _, f1_threshold, eer_threshold = run(
            name=threshold,
            predictions_folder=predictions_folder,
            steps=steps,
        )

        if (f1_threshold is not None) and (eer_threshold is not None):
            logger.info(f"Set --f1_threshold={f1_threshold:.5f}")
            logger.info(f"Set --eer_threshold={eer_threshold:.5f}")

    elif isinstance(threshold, float):
        f1_threshold = threshold
        eer_threshold = f1_threshold
    else:
        raise ValueError("Threshold value is neither an int nor a float")

    results_dict = {  # type: ignore
        "pred_data": defaultdict(dict),
        "fig_score": defaultdict(dict),
        "maxf1_threshold": defaultdict(dict),
        "post_eer_threshold": defaultdict(dict),
    }

    for k in dataloader.keys():
        if k.startswith("_"):
            logger.info(f"Skipping dataset '{k}' (not to be evaluated)")
            continue
        logger.info(f"Analyzing '{k}' set...")
        pred_data, fig_score, maxf1_threshold, post_eer_threshold = run(
            k,
            predictions_folder,
            f1_thresh=f1_threshold,
            eer_thresh=eer_threshold,
            steps=steps,
        )

        results_dict["pred_data"][k] = pred_data
        results_dict["fig_score"][k] = fig_score
        results_dict["maxf1_threshold"][k] = maxf1_threshold
        results_dict["post_eer_threshold"][k] = post_eer_threshold

    if output_folder is not None:
        output_scores = os.path.join(output_folder, "scores.pdf")

        if output_scores is not None:
            output_scores = os.path.realpath(output_scores)
            logger.info(f"Creating and saving scores at {output_scores}...")
            os.makedirs(os.path.dirname(output_scores), exist_ok=True)

            score_pdf = PdfPages(output_scores)

            for fig in results_dict["fig_score"].values():
                score_pdf.savefig(fig)
            score_pdf.close()

        data = {}
        for subset_name in dataloader.keys():
            data[subset_name] = {
                "df": results_dict["pred_data"][subset_name],
                "threshold": results_dict["post_eer_threshold"][threshold]
                if isinstance(threshold, str)
                else eer_threshold,
                "threshold_type": f"posteriori [{threshold}]"
                if isinstance(threshold, str)
                else "priori",
            }

        output_figure = os.path.join(output_folder, "plots.pdf")

        if output_figure is not None:
            output_figure = os.path.realpath(output_figure)
            logger.info(f"Creating and saving plots at {output_figure}...")
            os.makedirs(os.path.dirname(output_figure), exist_ok=True)
            pdf = PdfPages(output_figure)
            pdf.savefig(precision_recall_f1iso(data))
            pdf.savefig(roc_curve(data))
            pdf.close()

        output_table = os.path.join(output_folder, "table.txt")
        logger.info("Tabulating performance summary...")
        table = performance_table(data, "rst")
        click.echo(table)
        if output_table is not None:
            output_table = os.path.realpath(output_table)
            logger.info(f"Saving table at {output_table}...")
            os.makedirs(os.path.dirname(output_table), exist_ok=True)
            with open(output_table, "w") as f:
                f.write(table)
